-- == SRB2Randomizer: Prize randomization ==
-- Switches around "major" powerups/collectibles (monitors, tokens, emblems)

randomizer.cvars.monitors = CV_RegisterVar({
	name = "randomizer_monitors",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No monitor randomization.
		-- Use the original level's monitor layout.
		Replace = 1,
		-- Monitors will be completely replaced with new ones,
		-- from the list of all monitor types in the game.
		-- This option creates more item variety.
		Shuffle = 2,
		-- Monitors will switch spots with each other.
		-- This option is a lot tamer, and keeps the level's item balance in-tact.
	},
	defaultvalue = "Replace",
})

randomizer.cvars.emeralds = CV_RegisterVar({
	name = "randomizer_emeralds",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = {
		Off = 0,
		-- No randomization for Chaos Emeralds & Emerald Tokens.
		-- Use the original level's placement.
		On = 1,
		-- Emerald objects will swap with any Monitor spots.
		-- This option requires more exploration & map knowledge to get all Emeralds.
		Remove = 2,
		-- Remove Emerald objects from the game entirely. Their spots will be used to add new monitors.
		-- This option is specifically meant for maps that don't have unique Special Stages, but still have Emerald Tokens.
		-- Can also be set if you hate having Super Sonic or hate going to Special Stages.
	},
	defaultvalue = "On",
})

randomizer.cvars.emblems = CV_RegisterVar({
	name = "randomizer_emblems",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff,
	defaultvalue = "Off",
	-- Set whenever or not to swap emblems with monitor and token spots, if they're available.
})

local function generateMonitorTypesTable()
	local monitortypes = {}

	for i = 0,#mobjinfo-1
		if (mobjinfo[i].flags & MF_MONITOR)
		and (mobjinfo[i].doomednum != -1) -- make sure it's intended to be placeable
		and (G_GametypeHasTeams() or not (i == MT_REDRINGBOX or i == MT_BLUERINGBOX)) -- CTF boxes
		and ((multiplayer or netgame) or not (i == MT_MIXUPBOX or i == MT_RECYCLETV)) -- don't spawn recyclers or swaps in singleplayer
			local addamount = 2

			-- Weighted amounts based on type
			if (i == MT_SUPERRINGBOX)
				-- Rings should be a lot more frequent
				addamount = 10
			elseif (i == MT_MIXUPBOX)
			or (i == MT_RECYCLETV)
			or (i == MT_GRAVITYBOX)
				-- Weirdo items should be rarer
				addamount = 1
			end

			for j = 1,addamount
				table.insert(monitortypes, i)
			end
		end
	end

	return monitortypes
end

local function replaceMonitors()
	local monitortypes = generateMonitorTypesTable()

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.flags & MF_MONITOR)
			if (mo.tracer and mo.tracer.valid) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end

			local newtype = monitortypes[(randomizer.rv % #monitortypes) + 1]
			if (mo.type != newtype)
				mo.type = newtype
				mo.state = mobjinfo[newtype].spawnstate
			end

			--[[
			if (mo.spawnpoint and mo.spawnpoint.valid)
				mo.spawnpoint.type = mobjinfo[mo.type].doomednum
			end
			--]]

			randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
				+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
				+ (mo.floorz >> FRACBITS)
			)
		end
	end
end

local function shuffleMonitors()
	local monitors = {}
	local monitortypes = {}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.flags & MF_MONITOR)
			if (mo.tracer and mo.tracer.valid) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end

			table.insert(monitors, mo)
			table.insert(monitortypes, mo.type)
		end
	end

	if (#monitors > 1)
		for i = 1,#monitors
			if (#monitortypes <= 0)
				break
			end

			local mo = monitors[i]
			local typeindex = (randomizer.rv % #monitortypes) + 1
			local newtype = monitortypes[typeindex]

			if (mo.type != newtype)
				mo.type = newtype
				mo.state = mobjinfo[newtype].spawnstate
			end

			--[[
			if (mo.spawnpoint and mo.spawnpoint.valid)
				mo.spawnpoint.type = mobjinfo[mo.type].doomednum
			end
			--]]

			table.remove(monitortypes, typeindex)
			randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
				+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
				+ (mo.floorz >> FRACBITS)
			)
		end
	end
end

local function shuffleEmeralds()
	local monitors = {}
	local tokens = {}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.flags & MF_MONITOR)
			if (mo.tracer and mo.tracer.valid) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end
			table.insert(monitors, mo)
		elseif (mo.type == MT_EMMY)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			table.insert(tokens, mo)
		end
	end

	if (#tokens > 0 and #monitors > 0)
		for i = 1,#tokens
			local swap = (randomizer.rv % #monitors) + 1
			local token = tokens[i]
			local monitor = monitors[swap]

			randomizer.rv = abs($1 + (token.x >> FRACBITS) + (monitor.y >> FRACBITS)
				+ (token.z >> FRACBITS) - monitor.sprite - (token.ceilingz >> FRACBITS)
				+ (monitor.floorz >> FRACBITS)
			)

			-- I don't really care about remaking the tokenbits stuff so I'm teleporting the things around instead
			local newx = monitor.x
			local newy = monitor.y
			local newz = monitor.z
			local newflip = (monitor.flags2 & MF2_OBJECTFLIP)

			if (newflip)
				newz = $1 + monitor.height - token.height - 24*token.scale
			else
				newz = $1 + 24*token.scale
			end

			P_TeleportMove(monitor, token.x, token.y, token.z)
			if (token.flags2 & MF2_OBJECTFLIP)
				monitor.eflags = $1|MFE_VERTICALFLIP
				monitor.flags2 = $1|MF2_OBJECTFLIP
			else
				monitor.eflags = $1 & ~MFE_VERTICALFLIP
				monitor.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			P_TeleportMove(token, newx, newy, newz)
			if (newflip)
				token.eflags = $1|MFE_VERTICALFLIP
				token.flags2 = $1|MF2_OBJECTFLIP
			else
				token.eflags = $1 & ~MFE_VERTICALFLIP
				token.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			if (token.tracer and token.tracer.valid)
				if (newflip)
					newz = $1 + token.height - token.tracer.height
				end

				P_TeleportMove(token.tracer, newx, newy, newz)
				if (newflip)
					token.tracer.eflags = $1|MFE_VERTICALFLIP
					token.tracer.flags2 = $1|MF2_OBJECTFLIP
				else
					token.tracer.eflags = $1 & ~MFE_VERTICALFLIP
					token.tracer.flags2 = $1 & ~MF2_OBJECTFLIP
				end
			end

			table.remove(monitors, swap)
		end
	end
end

local function removeEmeralds()
	local monitortypes = generateMonitorTypesTable()
	local removelater = {}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.type == MT_EMMY)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			if (mo.tracer and mo.tracer.valid) -- token overlay
				P_RemoveMobj(mo.tracer)
			end

			local monitor = P_SpawnMobj(mo.x, mo.y, mo.z, monitortypes[(randomizer.rv % #monitortypes) + 1])
			if (mo.flags2 & MF2_OBJECTFLIP)
				monitor.eflags = $1|MFE_VERTICALFLIP
				monitor.flags2 = $1|MF2_OBJECTFLIP
			else
				monitor.eflags = $1 & ~MFE_VERTICALFLIP
				monitor.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			randomizer.rv = abs($1 + (mo.x >> FRACBITS) + (mo.y >> FRACBITS)
				+ (mo.z >> FRACBITS) - mo.sprite - (mo.ceilingz >> FRACBITS)
				+ (mo.floorz >> FRACBITS)
			)

			table.insert(removelater, mo)
		end
	end

	if (#removelater > 0)
		for i = 1,#removelater
			if (removelater[i] and removelater[i].valid)
				P_RemoveMobj(removelater[i])
			end
		end
	end
end

local function shuffleEmblems()
	local spots = {}
	local emblems = {}

	for mo in thinkers.iterate("mobj")
		if not (mo and mo.valid)
			continue
		end

		if (mo.flags & MF_MONITOR)
		or (mo.type == MT_EMMY)
		or (mo.type == MT_EMERALD1)
		or (mo.type == MT_EMERALD2)
		or (mo.type == MT_EMERALD3)
		or (mo.type == MT_EMERALD4)
		or (mo.type == MT_EMERALD5)
		or (mo.type == MT_EMERALD6)
		or (mo.type == MT_EMERALD7)
			if ((mo.flags & MF_MONITOR) and (mo.tracer and mo.tracer.valid)) -- 1up overlay
				P_RemoveMobj(mo.tracer)
			end
			table.insert(spots, mo)
		elseif (mo.type == MT_EMBLEM)
			table.insert(emblems, mo)
		end
	end

	if (#emblems > 0 and #spots > 0)
		for i = 1,#emblems
			local swap = (randomizer.rv % #spots) + 1
			local emblem = emblems[i]
			local othermo = spots[swap]

			randomizer.rv = abs($1 + (emblem.x >> FRACBITS) + (othermo.y >> FRACBITS)
				+ (emblem.z >> FRACBITS) - othermo.sprite - (emblem.ceilingz >> FRACBITS)
				+ (othermo.floorz >> FRACBITS)
			)

			local newx = othermo.x
			local newy = othermo.y
			local newz = othermo.z
			local newflip = (othermo.flags2 & MF2_OBJECTFLIP)

			if (newflip)
				newz = $1 + othermo.height - emblem.height
				if (othermo.flags & MF_MONITOR)
					newz = $1 - 24*emblem.scale
				end
			else
				if (othermo.flags & MF_MONITOR)
					newz = $1 + 24*emblem.scale
				end
			end

			P_TeleportMove(othermo, emblem.x, emblem.y, emblem.z)
			othermo.eflags = $1 & ~MFE_VERTICALFLIP
			othermo.flags2 = $1 & ~MF2_OBJECTFLIP

			if (othermo.type == MT_EMMY and (othermo.tracer and othermo.tracer.valid))
				P_TeleportMove(othermo.tracer, newx, newy, newz)
				othermo.tracer.eflags = $1 & ~MFE_VERTICALFLIP
				othermo.tracer.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			P_TeleportMove(emblem, newx, newy, newz)
			if (newflip)
				emblem.eflags = $1|MFE_VERTICALFLIP
				emblem.flags2 = $1|MF2_OBJECTFLIP
			else
				emblem.eflags = $1 & ~MFE_VERTICALFLIP
				emblem.flags2 = $1 & ~MF2_OBJECTFLIP
			end

			table.remove(spots, swap)
		end
	end
end

addHook("MapLoad", function()
	if (randomizer.seed == nil or randomizer.rv == nil)
		return
	end

	if (randomizer.cvars.monitors.value == 1)
		replaceMonitors()
	elseif (randomizer.cvars.monitors.value == 2)
		shuffleMonitors()
	end

	if (randomizer.cvars.emeralds.value == 1)
		shuffleEmeralds()
	elseif (randomizer.cvars.emeralds.value == 2)
		removeEmeralds()
	end

	if (randomizer.cvars.emblems.value == 1)
		shuffleEmblems()
	end
end)
